import pathlib


class PrettyReprMixin:
    """
    Mixin providing a common interface for
    __repr__ methods which represents classes
    in a very readable way.

    - How to use:
    Define which fields to include when printing the
    class or calling repr(some_object), by adding their
    names to the '__repr_fields__' tuple.

    This is just too handy!
    """
    __repr_fields__ = ()

    def __repr__(self):
        name = self.__class__.__name__
        attrs = [f"{i}={getattr(self, str(i))}" for i in self.__repr_fields__]
        return f"{name}({', '.join(attrs)})"


class PyttmanCliComplainerMixin:
    """
    Provides implementation classes a set of methods commonly
    used to complain to users when criteria aren't met
    """
    @staticmethod
    def complain_app_not_found(app_name: str | None) -> str | None:
        if app_name is None:
            return "Please provide a name for your app."
        if not pathlib.Path(app_name).exists():
            return f"- App '{app_name}' was not found here, " \
                   f"verify that a Pyttman app directory " \
                   f"named '{app_name}' exists."



class PyttmanProjectInvalidException(BaseException):
    """
    Exception class for internal use.
    Raise this exception in situations when a user
    project is incorrectly configured in whatever way.
    """
    pass


class TypeConversionFailed(BaseException):
    """
    Exception class for internal use.
    Raise this exception when a type conversion fails.
    """

    def __init__(self, from_type, to_type):
        message = f"Type '{from_type}' could not be converted to '{to_type}'."
        super().__init__(message)


class InvalidPyttmanObjectException(BaseException):
    """
    This error is raised when a user-implementation of a component within
    the Pyttman API is not correctly configured by the developer.
    """

    def __init__(self, message):
        super().__init__(message)


class ClientImproperlyConfiguredError(BaseException):
    """
    This error is raised when a Client is misconfigured in settings.py
    for a pyttman project.
    """

    def __init__(self, message):
        super().__init__(message)


class PyttmanPluginException(BaseException):
    """
    This error is raised when a plugin fails to execute one
    or several hooks
    """



import abc

from pyttman.core.containers import MessageMixin, Reply, ReplyStream
from pyttman.core.middleware.routing import AbstractMessageRouter


class BaseClient(abc.ABC):
    """
    Baseclass for Clients, for interfacing
    with users directly or through an API.
    """
    def __init__(self,
                 message_router: AbstractMessageRouter,
                 plugins: list,
                 **kwargs):
        self.message_router = message_router
        self.name = self.__class__.__name__
        self.plugins = plugins

        [setattr(self, k, v) for k, v in kwargs.items()]

        if not self.message_router:
            raise AttributeError("Pyttman Clients require an "
                                 "instance of AbstractMessageRouter "
                                 "subclass for routing messages to "
                                 "Intents")

    def __repr__(self):
        return f"{self.name}({vars(self)})"

    @abc.abstractmethod
    def run_client(self, *args, **kwargs) -> None:
        """
        Starts the main method for the client, opening
        a session to the front end with which it is
        associated with.
        """
        pass

    def reply_to_message(self, message: MessageMixin) -> Reply | ReplyStream:
        """
        Wrapper for calling the message router for a reply, whilst
        calling plugin methods in order.
        """
        return self.message_router.get_reply(message)



import inspect
from decimal import Decimal
from abc import ABC
from typing import Any, Sequence, Type

from pyttman.core.entity_parsing.entity import Entity

from pyttman.core.containers import MessageMixin, Message

from pyttman.core.entity_parsing.identifiers import IntegerIdentifier, \
    Identifier
from pyttman.core.entity_parsing.parsers import EntityFieldValueParser
from pyttman.core.exceptions import TypeConversionFailed, \
    InvalidPyttmanObjectException


class EntityFieldBase(EntityFieldValueParser, ABC):
    """
    EntityField classes make up the declarative API, to empower
    developers with a declarative and easy-to-use set of classes
    which specialize in identifying words or sentences in messages
    which match the given pattern.

    Not only to find the word(s) but also type-convert to a given
    datatype if a match is True.
    """
    default = None
    """
    Specify a default return value, if no match is found.
    """
    type_cls = None
    """
    The type_cls is the class which the value is converted to.
    For example: int, str, float, etc.
    """
    identifier_cls = None
    """
    Identifier class is a class which specializes in finding
    patterns in text. You can provide a custom Identifier class
    to further increase the granularity of the value you're looking for.
    """
    post_processor = None
    """
    Pre-processor is a callable which is called before the
    value is converted to the type_cls of the EntityField.
    You can specify any callable here, and it will be called
    with the value as its only argument.
    """

    def __init__(self,
                 identifier: Type[Identifier] | None = None,
                 default: Any = None,
                 post_processor: callable = None,
                 **kwargs):
        """
        :param as_list: If set to True combined with providing 'valid_strings',
                        the matched strings, if multiple, are stored as a list
                        rather than being a concatenated single string.

        :param identifier: Optional Identifier class. Provide an optional
               Identifier to further increase the granularity of the
               value you're looking for. Identifier classes can help in
               finding numbers, capitalized strings - or any other pattern
               by subclassing Identifier and defining your own regex pattern.
               A few common options are; CellphoneNumberIdentifier,
               CapitalizedIdentifier, and DateTimeStringIdentifier.
               You can read more about Identifier classes in the Pyttman
               documentation.
        """
        self.post_processor = post_processor
        if self.type_cls is None or inspect.isclass(self.type_cls) is False:
            raise InvalidPyttmanObjectException("All EntityField classes "
                                                "must define a 'type_cls', "
                                                "and it must be a class, "
                                                "not an object or primitive. "
                                                f"'{self.type_cls}' is not "
                                                f"a valid value for "
                                                f"'type_cls'.")

        _default_arg = default if default is not None else self.default
        super().__init__(identifier=identifier or self.identifier_cls,
                         default=_default_arg, **kwargs)

    def convert_value(self, value: Any) -> Any:
        """
        Try and convert the value passed, with the type associated
        with the class as 'type_cls'.

        :param value: Any
        :return: Any
        """
        if value is None or type(value) == self.type_cls:
            return value

        try:
            if self.as_list:
                if not isinstance(value, list):
                    converted_as_list = value.split()
                else:
                    converted_as_list = value

                for i, _ in enumerate(converted_as_list):
                    current_item = converted_as_list[i]
                    custom_mutation = self.before_conversion(current_item)
                    converted_value = self.type_cls(custom_mutation)
                    converted_as_list[i] = converted_value
                return converted_as_list
            else:
                custom_mutation = self.before_conversion(value)
                converted_value = self.type_cls(custom_mutation)
        except Exception as e:
            raise TypeConversionFailed(from_type=type(value),
                                       to_type=self.type_cls) from e

        try:
            if self.post_processor is not None and callable(self.post_processor):
                converted_value = self.post_processor(converted_value)
        except Exception as e:
            value_err = ValueError("The post_processor callable '"
                                   f"'{self.post_processor}' failed: {e}")
            raise TypeConversionFailed(from_type=type(value),
                                       to_type=self.type_cls) from value_err
        return converted_value

    def before_conversion(self, value: Any) -> str:
        """
        Perform the conversion from unknown type in 'value
        with the logic and constraints known only by the
        subclass with knowledge about its typecast.

        :param value: Any
        :return: Any
        """
        return value


class IntegerEntityField(EntityFieldBase):
    """
    IntegerEntityField classes specialize in finding numbers.
    The value output type from this EntityField is <int>.
    """
    type_cls = int
    identifier_cls = IntegerIdentifier

    def before_conversion(self, value: str) -> str:
        if isinstance(value, str):
            return "".join(i for i in value if i.isdigit())
        return value


class FloatEntityField(EntityFieldBase):
    """
    IntegerEntityField classes specialize in finding numbers.
    The value output type from this EntityField is <float>.
    """
    type_cls = float
    identifier_cls = IntegerIdentifier

    def before_conversion(self, value: str) -> str:
        converted = "".join(i for i in value if i.isdigit() or i in ".,")
        try:
            converted = converted.replace(",", ".")
        except AttributeError:
            pass
        return converted


class TextEntityField(EntityFieldBase):
    """
    TextEntityField classes specialize in finding text, of any kind.
    The value output type from this EntityField is <str>.
    """
    type_cls = str


class BoolEntityField(EntityFieldBase):
    """
    BoolEntityField will be set to True if a matching value is found
    in the Message. If no matching value is found, the value is set to
    False.
    The value output from this EntityField is <int>.
    """
    type_cls = bool
    default = False
    truncates_message_in_parsing = True

    def __init__(self, *args,
                 message_contains: Sequence[str] = None,
                 truncate_message_in_parsing: bool = True,
                 **kwargs):
        super().__init__(*args,
                         valid_strings=message_contains,
                         truncate_message_in_parsing=truncate_message_in_parsing,
                         **kwargs)

    def parse_message(self,
                      message: MessageMixin,
                      original_message_content: tuple[str],
                      memoization: dict = None) -> None:
        original_content = Message(original_message_content).lowered_content()
        self.value = Entity(value=self.default, is_fallback_default=True)
        if set(self.valid_strings).intersection(original_content):
            self.value = Entity(value=True)


class DecimalEntityField(FloatEntityField):
    """
    Converts floating point numbers to Decimal. Useful
    when dealing with currencies or in other areas where
    decimal precision is not trivial.
    """
    type_cls = Decimal
    default = Decimal(0)


if __name__ != "__main__":
    StringEntityField = TextEntityField
    StrEntityField = TextEntityField
    IntEntityField = IntegerEntityField



import code
import os
import re
import traceback
from pathlib import Path

import requests

from time import sleep
from pyttman.version import __version__ as pyttman_version
from pyttman.core.mixins import PyttmanCliComplainerMixin
from pyttman.core.decorators import LifeCycleHookType
from pyttman.core.entity_parsing.fields import TextEntityField
from pyttman.core.intent import Intent
from pyttman.tools.pyttmancli import TerraFormer, bootstrap_app
from pyttman.core.containers import (
    Message,
    ReplyStream,
    Reply
)


class ShellMode(Intent, PyttmanCliComplainerMixin):
    """
    Opens an interactive shell in their application,
    useful for debugging.
    """
    lead = ("shell",)
    example = "pyttman shell <app name>"
    help_string = "Opens a Python interactive shell with access to modules, " \
                  "app settings and the Pyttman 'app' object."
    app_name = TextEntityField(default="")

    def respond(self, message: Message) -> Reply | ReplyStream:
        app_name = message.entities["app_name"]
        if complaint := self.complain_app_not_found(app_name):
            return Reply(complaint)
        app = bootstrap_app(devmode=True, module=app_name)
        app.hooks.trigger(LifeCycleHookType.before_start)
        app.execute_plugins_before_app_start()
        global_variables = globals().copy()
        global_variables.update(locals())
        shell = code.InteractiveConsole(global_variables)
        shell.interact()
        app.execute_plugins_after_app_stop()
        return Reply("Exited shell")


class CreateNewApp(Intent, PyttmanCliComplainerMixin):
    """
    Create a new Pyttman app. The directory is terraformed
    and prepared with a template project.
    """
    app_name = TextEntityField(default="")
    lead = ("new",)
    trail = ("app",)
    ordered = True
    example = "pyttman new app <app name>"
    help_string = "Creates a new Pyttman app project in current directory " \
                  "from a template.\n" \
                  f"Example: {example}"

    def respond(self, message: Message) -> Reply | ReplyStream:
        num_retries = 3
        net_err = None
        app_name = message.entities["app_name"]
        terraformer = TerraFormer(app_name=app_name,
                                  url=self.storage["template_url"])

        print(f"- Creating project '{app_name}'.")
        print(f"- Downloading the latest official project "
              f"template from GitHub", end="\n")

        for i in range(num_retries):
            print(f"- Attempt [{i + 1}/{num_retries}]:", end=" ")
            try:
                terraformer.terraform()
            except FileExistsError as e:
                return Reply(str(e))

            except requests.exceptions.ConnectionError as e:
                print("failed.",
                      "Cannot connect to the server, "
                      "trying again in 5 seconds...",
                      end="\n",
                      sep=" ")
                sleep(5)
                net_err = e
                continue
            except Exception as e:
                print(f"errors occurred: '{e}'. See full traceback below.",
                      traceback.format_exc(), end="\n\n", sep="\n\n")
                raise e
            else:
                print("success!")
                break

        if net_err is not None:
            raise RuntimeError(f"The project could not be created.")
        return Reply(f"Wondering what to do next? Visit the "
                     f"Pyttman Wiki to follow our Get Started guide at "
                     f"https://github.com/dotchetter/Pyttman/wiki/Tutorial")


class RunAppInDevMode(Intent, PyttmanCliComplainerMixin):
    """
    Run a Pyttman app in dev mode. This sets "DEV_MODE"
    to True and opens a chat interface in the terminal.
    """
    app_name = TextEntityField(default="")
    fail_gracefully = True
    lead = ("dev",)
    example = "pyttman dev <app name>"
    help_string = "Run a Pyttman app in dev mode. Dev mode sets " \
                  "'pyttman.DEBUG' to True, enabling verbose outputs " \
                  "as defined in your app.\nThe app is started using " \
                  "a CliClient for you to start chatting with your app " \
                  "with minimal overhead.\n" \
                  f"Example: {example}"

    def respond(self, message: Message) -> Reply | ReplyStream:
        app_name = message.entities["app_name"]
        if complaint := self.complain_app_not_found(app_name):
            return Reply(complaint)
        try:
            app = bootstrap_app(devmode=True, module=app_name)
            app.hooks.trigger(LifeCycleHookType.before_start)
        except Exception as e:
            if self.fail_gracefully is False:
                raise e
            print(traceback.format_exc())
            return Reply("The app could not start due to issues with "
                         "bootstrapping, see traceback above.")
        app.execute_plugins_before_app_start()
        self.storage.put("app", app)
        self.storage.put("ready", True)
        return Reply(f"- Starting app '{app_name}' in dev mode...")


class RunAppInClientMode(Intent, PyttmanCliComplainerMixin):
    """
    Run a Pyttman in production mode. Client is configured
    settings.py.
    """
    fail_gracefully = True
    lead = ("runclient", "run")
    example = "pyttman run <app name>"
    help_string = "Run a Pyttman app in client mode. This is the " \
                  "standard production mode for Pyttman apps.\nThe " \
                  "app will be started using the client defined in " \
                  "settings.py under 'CLIENT'.\n" \
                  f"Example: {example}"

    app_name = TextEntityField(default="")

    def respond(self, message: Message) -> Reply | ReplyStream:
        app_name = message.entities["app_name"]
        if complaint := self.complain_app_not_found(app_name):
            return Reply(complaint)
        try:
            app = bootstrap_app(devmode=False, module=app_name)
            app.hooks.trigger(LifeCycleHookType.before_start)
        except Exception as e:
            if self.fail_gracefully is False:
                raise e
            print(traceback.format_exc())
            return Reply("The app could not start due to issues with "
                         "bootstrapping, see traceback above.")
        self.storage.put("app", app)
        self.storage.put("ready", True)
        app.execute_plugins_before_app_start()
        return Reply(f"- [Pyttman]: Starting app '{app_name}' in client mode")


class RunFile(Intent, PyttmanCliComplainerMixin):
    """
    Runs a single Python file within a Pyttman app context
    """
    fail_gracefully = True
    lead = ("runfile",)
    example = "pyttman runfile <app name> <file.py>"
    help_string = "Run a singe file within a Pyttman app context. " \
                  f"Example: {example}"

    app_name = TextEntityField(default="")
    script_file_name = TextEntityField(prefixes=(app_name,))

    def respond(self, message: Message) -> Reply | ReplyStream:
        app_name = message.entities["app_name"]
        if (script_path := message.entities["script_file_name"]) is None:
            return Reply("Filename must be entered, as a '.py' file. "
                         f"Example: {self.example}")
        if complaint := self.complain_app_not_found(app_name):
            return Reply(complaint)

        script_path = Path(script_path)
        app = bootstrap_app(devmode=True, module=app_name)
        app.hooks.trigger(LifeCycleHookType.before_start)
        global_variables = globals().copy()
        global_variables.update(locals())
        shell = code.InteractiveConsole(global_variables)
        app.execute_plugins_before_app_start()
        with open(script_path.as_posix(), "r") as f:
            # Set variable to indicate for the running script that it's main
            source = f.read()
            code_obj = compile(source, script_path.as_posix(), "exec")
            global_variables["__name__"] = "__main__"
            exec(code_obj, global_variables)
        return Reply(f"Script file executed successfully.")


class CreateNewAbilityIntent(Intent, PyttmanCliComplainerMixin):
    """
    Create a new Ability in a Pyttman application. A new directory
    and a 'ability.py' file is created for you in the app directory.
    """
    lead = ("new",)
    trail = ("ability",)
    ordered = True
    example = "pyttman new ability <ability name> app <app name>"
    help_string = "Create a new file with an Ability class as " \
                  "a template for new Ability classes for your app.\n" \
                  f"Example: {example}"

    app_name = TextEntityField()
    ability_name = TextEntityField(prefixes=(app_name,))

    def respond(self, message: Message) -> Reply | ReplyStream:
        files_to_create = ("__init__.py", "ability.py",
                           "intents.py", "models.py")
        ability_name = message.entities["ability_name"]
        app_name = message.entities["app_name"]
        if complaint := self.complain_app_not_found(app_name):
            return Reply(complaint)
        else:
            app_name = Path(app_name)

        abilities_parent_catalog = app_name / "abilities"
        ability_catalog = abilities_parent_catalog / ability_name

        if not abilities_parent_catalog.exists():
            os.mkdir(abilities_parent_catalog)
            init_file_path = abilities_parent_catalog / "__init__.py"
            with open(init_file_path, "w", encoding="utf-8") as f:
                # Empty file is OK
                pass

        if not ability_catalog.exists():
            os.mkdir(ability_catalog)
        else:
            return Reply("There's already an ability with that name.")

        for file in files_to_create:
            rel_path = ability_catalog / file
            with open(rel_path, "w", encoding="utf-8") as f:
                f.write("\n# Created by Pyttman ")
        return Reply(f"Created ability '{ability_name}'.")


class VersionInfo(Intent):
    """
    Returns the version info about this Pyttman installation.
    """
    lead = ("-v", "v",)
    example = "pyttman -V"
    help_string = "Returns the Pyttman build version. " \
                  f"Example: {example}"

    def respond(self, message: Message) -> Reply | ReplyStream:
        return Reply(pyttman_version)




"""
This file contains Intents and Ability classes which make out
the Pyttman cli tool, used to administer, bootstrap and create
Pyttman app for users of the framework.

Are you a Pyttman developer or a contributor and want to extend
the 'pyttman' terminal tool with new functionality?

Add Intents to the PyttmanCLI Ability and ensure that you
overload the 'description' and 'example' fields of the Intent
classes, since they provide essential data for users when
using the tool in the terminal.

"""
import argparse
import sys
import typing

from pyttman.core.containers import Message
from pyttman.core.middleware.routing import FirstMatchingRouter
from pyttman.tools.pyttmancli.terraforming import TerraFormer, bootstrap_app
from pyttman.tools.pyttmancli.terraforming import bootstrap_app, TerraFormer
from pyttman.tools.pyttmancli.ability import PyttmanCli


def run(argv=None, dev_args: typing.List = None):
    """
    This function utilizes the Pyttman framework itself,
    to administrate, bootstrap and create Pyttman apps.

    The method is designed to be implicitly called through
    the terminal shell, as argparse is used to extract
    values for parsing.

    For testing purposes if you're a Pyttman developer;
    provide your arguments as a single list of strings
    in the 'dev_args' kwarg instead of using 'argv'.

    This allows the Pyttman cli to work without actual
    terminal arguments.

    :param argv: Args from a terminal shell.
    :param dev_args: Optional developer args in a list of
                     strings, for unit testing the Pyttman CLI
                     or otherwise using it outside a terminal
                     shell. Defining this argument will automatically
                     disregard the 'argv' arguments
    :return: None
    """
    terminal_message = Message()
    pyttman_cli = PyttmanCli()
    dot = "\u2022"
    default_response = [pyttman_cli.description]
    default_response.extend([f"\n{dot} {i.example}{i.__doc__}"
                             for i in pyttman_cli.intents])
    default_response = str(" ").join(default_response) + "\n"

    router = FirstMatchingRouter(abilities=[pyttman_cli], help_keyword="help",
                                 intent_unknown_responses=[default_response])

    # Extract args from terminal shell
    if argv is None:
        argv = sys.argv[:]
        arg_parser = argparse.ArgumentParser(prog="Pyttman CLI",
                                             usage="%(prog)s command",
                                             add_help=False,
                                             allow_abbrev=False)
        arg_parser.add_argument("args", nargs="*")

        try:
            command = argv[1]
        except IndexError:
            command = ""

        options, args = arg_parser.parse_known_args(argv[2:])
        terminal_message = Message(command.split() + options.args)

    elif dev_args is not None:
        terminal_message = Message(dev_args)

    reply = router.get_reply(terminal_message)
    print(reply.as_str())

    if pyttman_cli.storage.get("ready"):
        pyttman_cli.run_application()



import concurrent.futures
import functools
from enum import Enum, auto
from functools import singledispatchmethod

from pyttman.core.exceptions import InvalidPyttmanObjectException
from pyttman.core.mixins import PrettyReprMixin


class LifeCycleHookType(Enum):
    """
    Enum class describing which lifecycle hooks are available
    for use in the Pyttman environment.
    """
    before_start = auto()


class LifecycleHookRepository(PrettyReprMixin):
    """
    Container class, encapsulating lifecycle hooks
    for a certain execution point in the lifetime
    of an application.
    """
    __repr_fields__ = ("repository",)

    def __init__(self):
        self.repository: dict = {k: [] for k in LifeCycleHookType}

    def _add_hook(self, hook_type: LifeCycleHookType, hook: callable):
        """
        Adds a lifecycle hook to the repository for the corresponding list
        under a certain lifecycle hook type.
        :raise InvalidPyttmanObjectException: Raised when an invalid
               LifeCycleHookType is referenced
        """
        try:
            self.repository[hook_type].append(hook)
        except ValueError:
            raise InvalidPyttmanObjectException(
                f"The lifecycle hook '{hook_type}' is not supported."
                f"Valid options are: {[i.name for i in LifeCycleHookType]}")

    @singledispatchmethod
    def run(self, *args, **kwargs):
        """
        Decorator.
        Decorate functions or methods in a Pyttman application to have them
        executed in certain points in time for a Pyttman app lifecycle.
        """
        raise NotImplementedError("Cannot directly invoke dispatch method")

    @run.register
    def _(self, hook_type: LifeCycleHookType, *args, **kwargs) -> callable:
        """
        Offers a decorator accepting the Enum 'LifeCycleHookType'
        to tell when the hook is going to be triggered.
        """
        def decorator(hook: callable):
            partial = functools.partial(hook, *args, **kwargs)
            self._add_hook(hook_type, partial)
        return decorator

    @run.register
    def _(self, hook_type: str, *args, **kwargs) -> callable:
        """
        Offers a decorator accepting strings to define the name
        of the LifeCycleHookType to tell when the hook is going to be
        triggered.
        """
        def decorator(hook: callable):
            hook_enum = None
            try:
                hook_enum = LifeCycleHookType[hook_type]
            except KeyError:
                pass  # self._add_hook will raise InvalidPyttmanObjectException
            partial = functools.partial(hook, *args, **kwargs)
            self._add_hook(hook_enum, partial)
        return decorator

    def trigger(self, hook_type: LifeCycleHookType):
        """
        Trigger all hooks under a certain hook key.
        Hooks are executed concurrently using a thread pool.
        """
        with concurrent.futures.ThreadPoolExecutor() as executor:
            [executor.submit(i) for i in self.repository.get(hook_type)]



"""
This module contains tools for developers contributing to Pyttman.

It helps with building the framework for testing and simplifies
deployment, and makes it easier to verify functionality of tools such
as Pyttman CLI and testing the framework using a local-only test-app
with the most recent build of the framework.
"""

